'''
Function:
    音乐下载器
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import sys
import copy
import json
import click
import threading
if __name__ == '__main__':
    from modules import *
    from __init__ import __version__
else:
    from .modules import *
    from .__init__ import __version__


'''basic info'''
BASICINFO = '''************************************************************
Function: 音乐下载器 V%s
Author: Charles
微信公众号: Charles的皮卡丘
操作帮助:
    输入r: 重新初始化程序(即返回主菜单)
    输入q: 退出程序
    下载多首歌曲: 选择想要下载的歌曲时,输入{1,2,5}可同时下载第1,2,5首歌
歌曲保存路径:
    当前路径下的%s文件夹内
************************************************************'''


'''音乐下载器'''
class musicdl():
    def __init__(self, configpath=None, config=None, **kwargs):
        assert configpath or config, 'configpath or config should be given'
        self.config = loadConfig(configpath) if config is None else config
        self.logger_handle = Logger(self.config['logfilepath'])
        self.initializeAllSources()
    '''非开发人员外部调用-手动输入版'''
    def run(self, target_srcs=None):
        while True:
            print(BASICINFO % (__version__, self.config.get('savedir')))
            # 音乐搜索
            user_input = self.dealInput('请输入歌曲搜索的关键词: ')
            target_srcs = [
                'kugou', 'kuwo', 'qqmusic', 'qianqian', 'fivesing',
                'netease', 'migu', 'joox', 'yiting',
            ] if target_srcs is None else target_srcs
            search_results = self.search(user_input, target_srcs)
            # 打印搜索结果
            title = ['序号', '歌手', '歌名', '大小', '时长', '专辑', '来源']
            items, records, idx = [], {}, 0
            for key, values in search_results.items():
                for value in values:
                    items.append([
                        colorize(str(idx), 'number'), 
                        colorize(value['singers'], 'singer'), 
                        value['songname'], 
                        value['filesize'] if value['ext'] != 'flac' else colorize(value['filesize'], 'flac'), 
                        value['duration'], 
                        value['album'], 
                        colorize(value['source'].upper(), 'highlight'),
                    ])
                    records.update({str(idx): value})
                    idx += 1
            printTable(title, items)
            # 音乐下载
            user_input = self.dealInput('请输入想要下载的音乐编号: ')
            need_download_numbers = user_input.replace(' ', '').split(',')
            songinfos = []
            for item in need_download_numbers:
                songinfo = records.get(item, '')
                if songinfo: songinfos.append(songinfo)
            self.download(songinfos)
    '''非开发人员外部调用-语音版'''
    def runbyspeech(self, target_srcs=None, baiduspeech_params=None):
        if baiduspeech_params is None:
            baiduspeech_params = dict(
                app_id='25419425', 
                api_key='fct6UMiQMLsp53MqXzp7AbKQ', 
                secret_key='p3wU9nPnfR7iBz2kM25sikN2ms0y84T3'
            )
        sr_api = SpeechRecognition(**baiduspeech_params)
        while True:
            print(BASICINFO % (__version__, self.config.get('savedir')))
            # 音乐搜索
            sr_api.synthesisspeak('请问您想听的歌曲名是什么呢?')
            target_srcs = ['migu'] if target_srcs is None else target_srcs
            sr_api.record()
            user_input = sr_api.recognition()
            search_results = self.search(user_input, target_srcs)
            # 音乐下载
            songinfos, songpaths = [], []
            for key, values in search_results.items():
                for value in values:
                    songinfos.append(value)
                    songpaths.append(os.path.join(value['savedir'], value['savename']+'.'+value['ext']))
            sr_api.synthesisspeak(f'共搜索到{len(songinfos)}首和{user_input}相关的歌曲, 将依次为您下载播放.')
            self.download(songinfos)
            # 音乐播放
            for songpath in songpaths:
                sr_api.synthesisspeak(audiopath=songpath)
    '''音乐搜索'''
    def search(self, keyword, target_srcs):
        self.logger_handle.info(f'正在搜索 {colorize(keyword, "highlight")} 来自 {colorize("|".join([c.upper() for c in target_srcs]), "highlight")}')
        def threadSearch(search_api, keyword, target_src, search_results):
            try:
                search_results.update({target_src: search_api(keyword)})
            except Exception as err:
                self.logger_handle.error(str(err), True)
                self.logger_handle.warning('无法在%s中搜索 >>>> %s' % (target_src, keyword))
        task_pool, search_results = [], {}
        for target_src in target_srcs:
            task = threading.Thread(
                target=threadSearch,
                args=(getattr(self, target_src).search, keyword, target_src, search_results)
            )
            task_pool.append(task)
            task.start()
        for task in task_pool:
            task.join()
        return search_results
    '''音乐下载'''
    def download(self, songinfos):
        for songinfo in songinfos:
            getattr(self, songinfo['source']).download([songinfo])
    '''初始化所有支持的搜索/下载源'''
    def initializeAllSources(self):
        supported_sources = {
            'kuwo': Kuwo, 'joox': Joox, 'migu': Migu, 'kugou': Kugou,
            'lizhi': Lizhi, 'yiting': YiTing, 'netease': Netease, 'qqmusic': QQMusic,
            'qianqian': Qianqian, 'fivesing': FiveSing, 'ximalaya': Ximalaya,
        }
        for key, value in supported_sources.items():
            setattr(self, key, value(copy.deepcopy(self.config), self.logger_handle))
        return supported_sources
    '''处理用户输入'''
    def dealInput(self, tip=''):
        user_input = input(tip)
        if user_input.lower() == 'q':
            self.logger_handle.info('ByeBye')
            sys.exit()
        elif user_input.lower() == 'r':
            self.initializeAllSources()
            self.run()
        else:
            return user_input
    '''str'''
    def __str__(self):
        return 'Welcome to use musicdl!\nYou can visit https://github.com/CharlesPikachu/musicdl for more details.'


'''cmd直接运行'''
@click.command()
@click.version_option()
@click.option('-k', '--keyword', default=None, help='搜索的歌曲关键字, 若不指定, 则进入musicdl终端版')
@click.option('-l', '--logfilepath', default='musicdl.log', help='日志文件保存的路径')
@click.option('-p', '--proxies', default='{}', help='设置的代理')
@click.option('-s', '--savedir', default='music', help='下载的音乐保存路径')
@click.option('-c', '--count', default='5', help='在各个平台搜索时的歌曲搜索数量')
@click.option('-t', '--targets', default=None, help='指定音乐搜索的平台, 例如"migu,joox"')
def musicdlcmd(keyword, logfilepath, proxies, savedir, count, targets):
    config = {
        'logfilepath': logfilepath,
        'proxies': json.loads(proxies),
        'savedir': savedir,
        'search_size_per_source': int(count),
    }
    target_srcs = [
        'kugou', 'kuwo', 'qqmusic', 'qianqian', 'fivesing',
        'netease', 'migu', 'joox', 'yiting',
    ] if targets is None else [src.strip() for src in targets.split(',')]
    dl_client = musicdl(config=config)
    if keyword is None:
        dl_client.run(target_srcs=target_srcs)
    else:
        print(dl_client)
        search_results = dl_client.search(keyword, target_srcs)
        # 打印搜索结果
        title = ['序号', '歌手', '歌名', '大小', '时长', '专辑', '来源']
        items, records, idx = [], {}, 0
        for key, values in search_results.items():
            for value in values:
                items.append([
                    colorize(str(idx), 'number'), 
                    colorize(value['singers'], 'singer'), 
                    value['songname'], 
                    value['filesize'] if value['ext'] != 'flac' else colorize(value['filesize'], 'flac'), 
                    value['duration'], 
                    value['album'], 
                    colorize(value['source'].upper(), 'highlight'),
                ])
                records.update({str(idx): value})
                idx += 1
        printTable(title, items)
        # 音乐下载
        user_input = dl_client.dealInput('请输入想要下载的音乐编号: ')
        need_download_numbers = user_input.replace(' ', '').split(',')
        songinfos = []
        for item in need_download_numbers:
            songinfo = records.get(item, '')
            if songinfo: songinfos.append(songinfo)
        dl_client.download(songinfos)


'''run'''
if __name__ == '__main__':
    import os
    rootdir = os.path.split(os.path.abspath(__file__))[0]
    dl_client = musicdl(os.path.join(rootdir, 'config.json'))
    dl_client.run()


'''initialize'''
from .downloader import Downloader
from .speech import SpeechRecognition
from .logger import Logger, printTable, colorize
from .misc import touchdir, seconds2hms, loadConfig, filterBadCharacter


'''initialize'''
from .utils import (
    touchdir, seconds2hms, loadConfig, filterBadCharacter, Logger, printTable, 
    SpeechRecognition, Downloader, colorize
)
from .sources import (
    Kuwo, Migu, Joox, Lizhi, Kugou, YiTing, Netease, QQMusic, Qianqian, FiveSing, Ximalaya
)


'''
Function:
    下载器类
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import requests
from .misc import touchdir
from .logger import colorize
from alive_progress import alive_bar


'''下载器类'''
class Downloader():
    def __init__(self, songinfo, session=None, **kwargs):
        self.songinfo = songinfo
        self.session = requests.Session() if session is None else session
        self.__setheaders(songinfo['source'])
    '''外部调用'''
    def start(self):
        songinfo, session, headers = self.songinfo, self.session, self.headers
        touchdir(songinfo['savedir'])
        with session.get(songinfo['download_url'], headers=headers, stream=True) as response:
            if response.status_code not in [200]: return False
            total_size, chunk_size, downloaded_size = int(response.headers['content-length']), songinfo.get('chunk_size', 1024), 0
            savepath = os.path.join(songinfo['savedir'], f"{songinfo['savename']}.{songinfo['ext']}")
            text, fp = colorize('[FileSize]: %0.2fMB/%0.2fMB', 'pink'), open(savepath, 'wb')
            with alive_bar(manual=True) as bar:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if not chunk: continue
                    fp.write(chunk)
                    downloaded_size += chunk_size
                    bar.text(text % (downloaded_size / 1024 / 1024, total_size / 1024 / 1024))
                    bar(min(downloaded_size / total_size, 1))
        return True
    '''设置请求头'''
    def __setheaders(self, source):
        self.netease_headers = {}
        self.qqmusic_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36',
            'Referer': 'http://y.qq.com',
        }
        self.migu_headers = {
            'User-Agent': 'Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.75 Mobile Safari/537.36',
            'Referer': 'https://m.music.migu.cn/', 
        }
        self.baiduFlac_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36',
            'Referer': 'http://music.baidu.com/',
        }
        if hasattr(self, f'{source}_headers'):
            self.headers = getattr(self, f'{source}_headers')
        else:
            self.headers = {
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36'
            }


'''
Function:
    喜马拉雅下载: https://www.ximalaya.com/
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import time
import requests
from .base import Base
from ..utils import seconds2hms, filterBadCharacter


'''喜马拉雅下载类'''
class Ximalaya(Base):
    def __init__(self, config, logger_handle, **kwargs):
        super(Ximalaya, self).__init__(config, logger_handle, **kwargs)
        self.source = 'ximalaya'
        self.__initialize()
    '''歌曲搜索'''
    def search(self, keyword, disable_print=True):
        if not disable_print: self.logger_handle.info('正在%s中搜索 >>>> %s' % (self.source, keyword))
        cfg = self.config.copy()
        params = {
            'core': 'all',
            'kw': keyword,
            'spellchecker': 'true',
            'device': 'iPhone',
            'live': 'true',
        }
        response = self.session.get(self.search_url, params=params, headers=self.headers)
        all_items = response.json()['data']['track']['docs']
        songinfos = []
        for item in all_items:
            headers = self.headers.copy()
            headers['Referer'] = f'https://www.ximalaya.com/sound/{item["id"]}'
            response = self.session.get(self.songinfo_url.format(str(item['id'])), headers=headers)
            if response.json()['ret'] not in [200]: continue
            download_url = response.json()['data'].get('src', '')
            if not download_url: continue
            songinfo = {
                'source': self.source,
                'songid': str(item['id']),
                'singers': filterBadCharacter(item.get('nickname', '-')),
                'album': filterBadCharacter(item.get('albumTitle', '-')),
                'songname': filterBadCharacter(item.get('title', '-')),
                'savedir': cfg['savedir'],
                'savename': filterBadCharacter(item.get('title', f'{keyword}_{int(time.time())}')),
                'download_url': download_url,
                'lyric': '',
                'filesize': '-',
                'ext': download_url.split('.')[-1],
                'duration': seconds2hms(item.get('duration'))
            }
            songinfos.append(songinfo)
            if len(songinfos) == cfg['search_size_per_source']: break
        return songinfos
    '''初始化'''
    def __initialize(self):
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.63 Safari/537.36',
            'Referer': 'https://www.ximalaya.com/search/'
        }
        self.search_url = 'https://www.ximalaya.com/revision/search/main'
        self.songinfo_url = 'https://www.ximalaya.com/revision/play/v1/audio?id={}&ptype=1'